import java.io.FileWriter;

import cfg.BasicBlock;

import opt.Optimization;
import opt.Optimizer;
import opt.ssa.SSAConversionFactory;
import parser.*;
import semant.*;
import codegenmips.*;
import util.*;
import ast.*;

/**
 * Main class that runs the Bantam compiler Constructs and runs each phase of
 * the compiler (lexing, parsing, semantic analysis, and code generation).
 * */
public class Main {
  /** Array for holding each input file name */
  private static String[] inFiles;
  /** String that holds the output file name - "out.s" by default */
  private static String outFile = "out.s";
  /** Optimization level (0 means off) */
  private static int opt = 0;

  /** Prints out a usage message to the screen */
  private static void showHelp() {
    System.err
        .println("Usage: bantamc [-h] [-o <output_file>] [-opt <num>] <input_files>");
    System.exit(1);
  }

  /**
   * Processes the commandline flags, setting appropriate variables
   * 
   * @param args
   *          list of commandline arguments
   * */
  private static void processFlags(String[] args) {
    // initialize inFiles to size of args, will probably be smaller, but args
    // length gives upper bound
    inFiles = new String[args.length];
    // cnt represents the number of input files found - initialize to 0
    int cnt = 0;

    // if no arguments then call showHelp (which eventually exits)
    if (args.length == 0) showHelp();

    // otherwise inspect the arguments
    for (int i = 0; i < args.length; i++) {
      // if '-h' or help then call showHelp (which eventually exits)
      if (args[i].equals("-h"))
        showHelp();
      
      // if -opt then turn on optimization
      else if (args[i].equals("-opt")) {
        // check if no further arguments
        if (i == args.length - 1) {
          // if not, then print error message and call showHelp() (which
          // eventually exits)
          System.err
              .println("Usage error: must specify an optimization level with -opt");
          showHelp();
        }
        i++;

        boolean badLevel = true;
        try {
          opt = Integer.parseInt(args[i]);
          if (opt >= 0 && opt <= 4) badLevel = false;
        } catch (NumberFormatException e) {}

        if (badLevel) {
          // if not, then print error message and call showHelp() (which
          // eventually exits)
          System.err
              .println("Usage error: optimization level must be integer from 0-4");
          showHelp();
        }
      }

      // if -o is set then user is specifying the output file
      else if (args[i].equals("-o")) {
        // check if no further arguments
        if (i == args.length - 1) {
          // if not, then print error message and call showHelp() (which
          // eventually exits)
          System.err
              .println("Usage error: must specify an output file with -o");
          showHelp();
        }
        i++;
        // otherwise set outFile
        outFile = args[i];
      }

      // any other arguments must be input files

      // check if argument ends in .btm
      else if (args[i].length() >= 5
          && args[i].substring(args[i].length() - 4).equals(".btm")) {
        // if so then set next entry in inFiles
        inFiles[cnt++] = args[i];
      }

      else {
        // if we get to here then we have an illegal argument
        // (we treat this as a bad input file name)
        System.err.println("Usage error: bad input file name: " + args[i]);
        System.err.println("             file names must end with '.btm'");
        showHelp();
      }
    }

    // final check: make sure at least one input file was specified
    if (cnt == 0) {
      System.err.println("Usage error: must specify some input files");
      showHelp();
    }
    
    // finally, resize inFiles to the number of specified input files (i.e.,
    // cnt)
    String[] tmp = inFiles;
    inFiles = new String[cnt];
    for (int i = 0; i < cnt; i++)
      inFiles[i] = tmp[i];
  }

  /**
   * Main method, which drives compilation builds and runs each phase of the
   * compiler
   * 
   * @param args
   *          list of commandline arguments
   * */
  public static void main(String[] args) {
    // process flags
    processFlags(args);

    try {
      // parsing
      Parser parser = new Parser(inFiles, false);
      Program program = parser.parse();
      
      // semantic analysis
      SemanticAnalyzer semanticAnalyzer = new SemanticAnalyzer(program, false);
      ClassTreeNode classTree = semanticAnalyzer.analyze();
      
      // optimization
      final Optimizer optimizer = new Optimizer(classTree, false);
      optimizer.buildCFG();
      
      // convert to SSA form
      final Optimization cfg2ssa = SSAConversionFactory.createConverter();
      optimizer.optimize(cfg2ssa);

      // optimization passes (if specified via -opt flag)
      if (opt > 0) {
        // ... perform some optimizations according to opt level ...
        
        // basic example: 
        // 1) implement Optimization interface (in separate file, just 
        //    for illustration purposes here!)
        Optimization myOpt = 
        new Optimization() {
          
          @Override
          public BasicBlock transform() {
            // return the entry block of the CFG after transformation
            // (or new entry block, if CFG shall be replaced)
            return entryBB;
          }
          
          @Override
          public void setVerbose(boolean arg0) {}
          
          @Override
          public void setMethodName(String arg0) {}
          
          @Override
          public void setEntryBlock(BasicBlock arg0) {
            this.entryBB = arg0;
          }
          
          @Override
          public boolean isVerbose() { return false; }
          
          @Override
          public void analyze() {
            System.err.println("Test");
          }
          
          BasicBlock entryBB;
        };
        
        // 2) apply optimization
        optimizer.optimize(myOpt);
      }
      
      // convert back from SSA form
      final Optimization ssa2cfg = SSAConversionFactory.createDeconverter();
      optimizer.optimize(ssa2cfg);
      
      // code generation:
      // use this to get the default generator (faster compilation)
      final CodeGenerator codeGenerator = CodeGeneratorFactory.createCodeGeneratorMIPS(classTree);
      // use this to get the register allocating generator (faster code)
      /*final CodeGenerator codeGenerator = 
          CodeGeneratorFactory.createCodeGeneratorMIPS_RA(optimizer, classTree);*/
      codeGenerator.generate();
      
      // write to file
      try {
        final FileWriter fw = new FileWriter(outFile);
        fw.append(codeGenerator.getCode());
        fw.close();
      }
      catch (Throwable ex) {
        System.err.println ("Could not write to file '" + outFile + "': " + 
          ex.toString());
      }
    }
    catch (Exception e) {
      e.printStackTrace();
      System.err
          .println("Internal error within compiler: stopping compilation");
      System.exit(1);
    }
  }
}
